"""Random kernel classes for use with the RandomKernel layers."""
import numpy as np
import tensorflow as tf

from aboleth.random import seedgen
from aboleth.distributions import Normal, norm_posterior, kl_qp


#
# Random Fourier Kernels
#

class ShiftInvariant:
    """Abstract base class for shift invariant kernel approximations.

    Parameters
    ----------
    lenscale : float, ndarray, Tensor, Variable
        the length scales of the shift invariant kernel, this can be a scalar
        for an isotropic kernel, or a vector of shape (input_dim, 1) for an
        automatic relevance detection (ARD) kernel. If you wish to learn this
        parameter, make it a Variable (or ``ab.pos(tf.Variable(...))`` to keep
        it positively constrained).

    """

    def __init__(self, lenscale=1.0):
        """Constuct a shift invariant kernel object."""
        self.lenscale = lenscale

    def weights(self, input_dim, n_features):
        """Generate the random fourier weights for this kernel.

        Parameters
        ----------
        input_dim : int
            the input dimension to this layer.
        n_features : int
            the number of unique random features, the actual output dimension
            of this layer will be ``2 * n_features``.

        Returns
        -------
        P : ndarray
            the random weights of the fourier features of shape
            ``(input_dim, n_features)``.
        KL : Tensor, float
            the KL penalty associated with the parameters in this kernel.

        """
        raise NotImplementedError("Abstract base class for shift invariant"
                                  " kernels!")


class RBF(ShiftInvariant):
    """Radial basis kernel approximation.

    Parameters
    ----------
    lenscale : float, ndarray, Tensor, Variable
        the length scales of the shift invariant kernel, this can be a scalar
        for an isotropic kernel, or a vector of shape (input_dim, 1) for an
        automatic relevance detection (ARD) kernel. If you wish to learn this
        parameter, make it a Variable (or ``ab.pos(tf.Variable(...))`` to keep
        it positively constrained).

    """

    def weights(self, input_dim, n_features):
        """Generate the random fourier weights for this kernel.

        Parameters
        ----------
        input_dim : int
            the input dimension to this layer.
        n_features : int
            the number of unique random features, the actual output dimension
            of this layer will be ``2 * n_features``.

        Returns
        -------
        P : ndarray
            the random weights of the fourier features of shape
            ``(input_dim, n_features)``.
        KL : Tensor, float
            the KL penalty associated with the parameters in this kernel (0.0).

        """
        rand = np.random.RandomState(next(seedgen))
        e = rand.randn(input_dim, n_features).astype(np.float32)
        P = e / self.lenscale
        return P, 0.


class RBFVariational(ShiftInvariant):
    """Variational Radial basis kernel approximation.

    This kernel is similar to the RBF kernel, however we learn an independant
    Gaussian posterior distribution over the kernel weights to sample from.

    Parameters
    ----------
    lenscale : float, ndarray, Tensor, Variable
        the length scales of the shift invariant kernel, this can be a scalar
        for an isotropic kernel, or a vector of shape (input_dim, 1) for an
        automatic relevance detection (ARD) kernel. If you wish to learn this
        parameter, make it a Variable (or ``ab.pos(tf.Variable(...))`` to keep
        it positively constrained).
    lenscale_posterior : float, ndarray, optional
        the *initial* value for the posterior length scale, this can be a
        scalar or vector (different initial value per input dimension). If this
        is left as None, it will be set to ``sqrt(1 / input_dim)`` (this is
        similar to the 'auto' setting for a scikit learn SVM with a RBF
        kernel).

    """

    def __init__(self, lenscale=1.0, lenscale_posterior=None):
        """Constuct an instance of the RBFVariational kernel."""
        super().__init__(lenscale)
        self.lenscale_post = lenscale_posterior

    def weights(self, input_dim, n_features):
        """Generate the random fourier weights for this kernel.

        Parameters
        ----------
        input_dim : int
            the input dimension to this layer.
        n_features : int
            the number of unique random features, the actual output dimension
            of this layer will be ``2 * n_features``.

        Returns
        -------
        P : ndarray
            the random weights of the fourier features of shape
            ``(input_dim, n_features)``.
        KL : Tensor, float
            the KL penalty associated with the parameters in this kernel.

        """
        dim = (input_dim, n_features)

        # Setup the prior, lenscale may be a variable, so dont use prior_normal
        pP = Normal(mu=tf.zeros(dim), var=self.__len2var(self.lenscale))

        # Initialise the posterior
        if self.lenscale_post is None:
            self.lenscale_post = np.sqrt(1 / input_dim)
        qP = norm_posterior(dim=dim, var0=self.__len2var(self.lenscale_post))

        KL = kl_qp(qP, pP)

        # We implement the VAR-FIXED method here from Cutajar et. al 2017, so
        # we pre-generate and fix the standard normal samples
        rand = np.random.RandomState(next(seedgen))
        e = rand.randn(input_dim, n_features).astype(np.float32)
        P = qP.sample(e)

        return P, KL

    @staticmethod
    def __len2var(lenscale):
        var = tf.to_float(1. / lenscale**2)
        return var


class Matern(ShiftInvariant):
    """Matern kernel approximation.

    Parameters
    ----------
    lenscale : float, ndarray, Tensor, Variable
        the length scales of the shift invariant kernel, this can be a scalar
        for an isotropic kernel, or a vector of shape (input_dim, 1) for an
        automatic relevance detection (ARD) kernel. If you wish to learn this
        parameter, make it a Variable (or ``ab.pos(tf.Variable(...))`` to keep
        it positively constrained).
    p : int
        a zero or positive integer specifying the number of the Matern kernel,
        e.g. ``p == 0`` results int a Matern 1/2 kernel, ``p == 1``  results in
        the Matern 3/2 kernel etc.

    """

    def __init__(self, lenscale=1.0, p=1):
        """Constuct a Matern kernel object."""
        super().__init__(lenscale)
        assert isinstance(p, int) and p >= 0
        self.p = p

    def weights(self, input_dim, n_features):
        """Generate the random fourier weights for this kernel.

        Parameters
        ----------
        input_dim : int
            the input dimension to this layer.
        n_features : int
            the number of unique random features, the actual output dimension
            of this layer will be ``2 * n_features``.

        Returns
        -------
        P : ndarray
            the random weights of the fourier features of shape
            ``(input_dim, n_features)``.
        KL : Tensor, float
            the KL penalty associated with the parameters in this kernel (0.0).

        """
        # p is the matern number (v = p + .5) and the two is a transformation
        # of variables between Rasmussen 2006 p84 and the CF of a Multivariate
        # Student t (see wikipedia). Also see "A Note on the Characteristic
        # Function of Multivariate t Distribution":
        #   http://ocean.kisti.re.kr/downfile/volume/kss/GCGHC8/2014/v21n1/
        #   GCGHC8_2014_v21n1_81.pdf
        # To sample from a m.v. t we use the formula
        # from wikipedia, x = y * np.sqrt(df / u) where y ~ norm(0, I),
        # u ~ chi2(df), then x ~ mvt(0, I, df)
        df = 2 * (self.p + 0.5)
        rand = np.random.RandomState(next(seedgen))
        y = rand.randn(input_dim, n_features)
        u = rand.chisquare(df, size=(n_features,))
        P = (y * np.sqrt(df / u)).astype(np.float32) / self.lenscale
        return P, 0.



"""Test the layers module."""

import numpy as np
import tensorflow as tf

import aboleth as ab
from aboleth.losses import _sum_likelihood


def test_elbo_likelihood(make_graph):
    """Test for expected output dimensions from loss ."""
    x, y, N, X_, Y_, N_, like, layers = make_graph
    Net, kl = layers(X=X_)

    # Now test with likelihood weights
    loss = ab.elbo(Net, Y_, N, kl, like)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()

        L = loss.eval(feed_dict={X_: x, Y_: y, N_: float(N)})
        assert np.isscalar(L)


def test_map_likelihood(make_graph):
    """Test for expected output dimensions from deepnet."""
    x, y, _, _, _, _, like, layers = make_graph
    Net, reg = layers(X=x.astype(np.float32))

    # Now test with likelihood weights
    loss = ab.max_posterior(Net, y.astype(np.float32),  reg, like)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()

        L = loss.eval()
        assert np.isscalar(L)


def test_sum_likelihood():
    """Test we can do weighted sums of likelihoods."""
    like = ab.likelihoods.Bernoulli()
    N = 5
    Net = np.ones(N) * .5
    Y = np.ones(N)

    def weight_fn(Y):
        return Y * np.arange(N)

    unweighted = _sum_likelihood(Y, Net, like, None)
    value = _sum_likelihood(Y, Net, like, np.arange(N))
    call = _sum_likelihood(Y, Net, like, weight_fn)

    tc = tf.test.TestCase()
    with tc.test_session():
        sumll = unweighted.eval()
        assert np.allclose(sumll, np.log(0.5) * N)

        sumll = value.eval()
        assert np.allclose(sumll, np.sum(np.log(0.5) * np.arange(N)))

        sumll = call.eval()
        assert np.allclose(sumll, np.sum(np.log(0.5) * np.arange(N)))



"""Network loss functions."""
import tensorflow as tf


def elbo(Net, Y, N, KL, likelihood, like_weights=None):
    """Build the evidence lower bound loss for a neural net.

    Parameters
    ----------
    Net : ndarray, Tensor
        the neural net featues of shape (n_samples, N, tasks).
    Y : ndarray, Tensor
        the targets of shape (N, tasks).
    N : int, Tensor
        the total size of the dataset (i.e. number of observations).
    likelihood : Tensor
        the likelihood model to use on the output of the last layer of the
        neural net, see the :ref:`likelihoods` module.
    like_weights : callable, ndarray, Tensor
        weights to apply to each observation in the expected log likelihood.
        This should be an array of shape (N, 1) or can be called as
        ``like_weights(Y)`` and should return a (N, 1) array.

    Returns
    -------
    nelbo : Tensor
        the loss function of the Bayesian neural net (negative ELBO).

    """
    rank = len(Net.shape)
    assert rank > 2, "We need a Tensor of at least rank 3 for Bayesian models!"

    B = N / tf.to_float(tf.shape(Net)[1])  # Batch amplification factor
    n_samples = tf.to_float(Net.shape[0])  # averaging over samples

    # Just mean over samps for expected log-likelihood
    ELL = _sum_likelihood(Y, Net, likelihood, like_weights) / n_samples

    # negative ELBO is batch weighted ELL and KL
    nELBO = - B * ELL + KL

    return nELBO


def max_posterior(Net, Y, regulariser, likelihood, like_weights=None,
                  first_axis_is_obs=True):
    """Build maximum a-posteriori (MAP) loss for a neural net.

    Parameters
    ----------
    Net : ndarray, Tensor
        the neural net featues of shape (N, tasks) or (n_samples, N, tasks).
    Y : ndarray, Tensor
        the targets of shape (N, tasks).
    likelihood : Tensor
        the likelihood model to use on the output of the last layer of the
        neural net, see the :ref:`likelihoods` module.
    like_weights : callable, ndarray, Tensor
        weights to apply to each observation in the expected log likelihood.
        This should be an array of shape (N, 1) or can be called as
        ``like_weights(Y)`` and should return a (N, 1) array.
    first_axis_is_obs : bool
        indicates if the first axis indexes the observations/data or not. This
        will be True if ``Net`` is of shape (N, tasks) or False if ``Net`` is
        of shape (n_samples, N, tasks).

    Returns
    -------
    map : Tensor
        the loss function of the MAP neural net.

    """
    # Get the batch size to average the likelihood over
    M = tf.to_float(tf.shape(Net)[0 if first_axis_is_obs else 1])

    # Average likelihood for batch
    AVLL = _sum_likelihood(Y, Net, likelihood, like_weights) / M

    # MAP objective
    MAP = - AVLL + regulariser

    return MAP


#
# Private module utilities
#

def _sum_likelihood(Y, Net, likelihood, like_weights):
    """Sum the log-likelihood of the Y's under the model."""
    like = likelihood(Y, Net)
    if callable(like_weights):
        like *= like_weights(Y)
    elif like_weights is not None:
        like *= like_weights

    sumlike = tf.reduce_sum(like)

    return sumlike



"""Layers that impute missing data."""
import tensorflow as tf

from aboleth.baselayers import MultiLayer
from aboleth.distributions import Normal
from aboleth.random import seedgen
from aboleth.util import pos


class ImputeOp(MultiLayer):
    r"""Abstract Base Impute operation. These specialise MultiLayers.

    They expect a data InputLayer and a mask InputLayer. They return layers in
    which the masked values have been imputed.

    Parameters
    ----------
    datalayer : callable
        A layer that returns a data tensor. Must be of form ``f(**kwargs)``.
    masklayer : callable
        A layer that returns a boolean mask tensor where True values are
        masked. Must be of form ``f(**kwargs)``.

    """

    def __init__(self, datalayer, masklayer):
        """Construct and instance of an ImputeOp operation."""
        self.datalayer = datalayer
        self.masklayer = masklayer

    def _build(self, **kwargs):
        """Build an impute operation graph."""
        X_ND, loss1 = self.datalayer(**kwargs)
        M, loss2 = self.masklayer(**kwargs)

        self._check_rank(X_ND)
        self._set_mask(M)

        # Extra build/initialisation here
        self._initialise_variables(X_ND)

        # Map over the samples layer
        Net = tf.map_fn(self._impute2D, X_ND)

        loss = tf.add(loss1, loss2)
        return Net, loss

    def _impute2D(self, X_2D):
        r"""Impute a rank 2 tensor.

        This function is mapped over the rank 3 data tensors, additionally it
        has access to two properties:
        - ``self.missing_ind`` a row and column index of the missing data
        - ``self.real_val_mask`` a tf.float32 mask of the non missing values

        Parameters
        ----------
        X_2D : Tensor
            a rank 2 Tensor with missing data

        Returns
        -------
        X_imputed : Tensor
            a rank 2 Tensor with imputed data

        """
        raise NotImplementedError("Abstract base class for imputation ops!")
        X_imputed = None  # You imputation implementation
        return X_imputed

    def _check_rank(self, X):
        """Check the rank of the input tensors."""
        rank = len(X.shape)
        assert rank == 3

    def _set_mask(self, M):
        """Create Tensor Masks."""
        # Identify indices of the missing datapoints
        self.missing_ind = tf.where(M)
        self.real_val_mask = tf.cast(tf.logical_not(M), tf.float32)

    def _initialise_variables(self, X):
        """Optional extra build stage."""
        pass


class MeanImpute(ImputeOp):
    r"""Impute the missing values using the stochastic mean of their column.

    Takes two layers, one the returns a data tensor and the other returns a
    mask layer. Returns a layer that returns a tensor in which the masked
    values have been imputed as the column means calculated from the batch.

    Parameters
    ----------
    datalayer : callable
        A layer that returns a data tensor. Must be of form ``f(**kwargs)``.

    masklayer : callable
        A layer that returns a boolean mask tensor where True values are
        masked. Must be of form ``f(**kwargs)``.

    """

    def _impute2D(self, X_2D):
        r"""Mean impute a rank 2 tensor."""
        # Fill zeros in for missing data initially
        data_zeroed_missing_tf = X_2D * self.real_val_mask

        # Sum the real values in each column
        col_tot = tf.reduce_sum(data_zeroed_missing_tf, 0)

        # Divide column totals by the number of non-nan values
        num_values_col = tf.reduce_sum(self.real_val_mask, 0)
        num_values_col = tf.maximum(num_values_col,
                                    tf.ones(tf.shape(num_values_col)))
        col_nan_means = tf.div(col_tot, num_values_col)

        # Make an vector of the impute values for each missing point
        imputed_vals = tf.gather(col_nan_means, self.missing_ind[:, 1])

        # Fill the imputed values into the data tensor of zeros
        shape = tf.cast(tf.shape(data_zeroed_missing_tf), dtype=tf.int64)
        missing_imputed = tf.scatter_nd(self.missing_ind, imputed_vals, shape)

        X_with_impute = data_zeroed_missing_tf + missing_imputed

        return X_with_impute


class FixedNormalImpute(ImputeOp):
    r"""Impute the missing values using marginal Gaussians over each column.

    Takes two layers, one the returns a data tensor and the other returns a
    mask layer. Creates a layer that returns a tensor in which the masked
    values have been imputed as random draws from the marginal Gaussians.

    Parameters
    ----------
    datalayer : callable
        A layer that returns a data tensor. Must be of form ``f(**kwargs)``.
    masklayer : callable
        A layer that returns a boolean mask tensor where True values are
        masked. Must be of form ``f(**kwargs)``.
    mu_array : array-like
        A list of the global mean values of each dat column
    var_array : array-like
        A list of the global variance of each data column

    """

    def __init__(self, datalayer, masklayer, mu_array, var_array):
        """Construct and instance of a RandomGaussImpute operation."""
        super().__init__(datalayer, masklayer)
        self.normal_array = [Normal(m, v) for m, v in zip(mu_array, var_array)]

    def _impute2D(self, X_2D):
        r"""Randomly impute a rank 2 tensor."""
        # Fill zeros in for missing data initially
        data_zeroed_missing_tf = X_2D * self.real_val_mask

        # Divide column totals by the number of non-nan values
        col_draws = [n.sample() for n in self.normal_array]
        # Make an vector of the impute values for each missing point
        imputed_vals = tf.gather(col_draws, self.missing_ind[:, 1])

        # Fill the imputed values into the data tensor of zeros
        shape = tf.cast(tf.shape(data_zeroed_missing_tf), dtype=tf.int64)
        missing_imputed = tf.scatter_nd(self.missing_ind, imputed_vals, shape)

        X_with_impute = data_zeroed_missing_tf + missing_imputed

        return X_with_impute


class LearnedScalarImpute(ImputeOp):
    r"""Impute the missing values using learnt scalar for each column.

    Takes two layers, one the returns a data tensor and the other returns a
    mask layer. Creates a layer that returns a tensor in which the masked
    values have been imputed with a learned scalar value per colum.

    Parameters
    ----------
    datalayer : callable
        A layer that returns a data tensor. Must be an InputLayer.
    masklayer : callable
        A layer that returns a boolean mask tensor where True values are
        masked. Must be an InputLayer.

    """

    def __init__(self, datalayer, masklayer):
        r"""Construct and instance of a VarScalarImpute operation."""
        super().__init__(datalayer, masklayer)

    def _initialise_variables(self, X):
        """Initialise the impute variables."""
        datadim = int(X.shape[2])
        self.impute_scalars = tf.Variable(
            tf.random_normal(shape=(1, datadim), seed=next(seedgen)),
            name="impute_scalars"
        )

    def _impute2D(self, X_2D):
        r"""Randomly impute a rank 2 tensor.

        Parameters
        ----------
        X_2D : Tensor
            a rank 2 Tensor with missing data
        scalars : Tensor 1 x D
            these values are filled into the missing elements (per column)

        Returns
        -------
        X_imputed : Tensor
            a rank 2 Tensor with imputed data

        """
        # Fill zeros in for missing data initially
        data_zeroed_missing = X_2D * self.real_val_mask

        # Make an vector of the impute values for each missing point
        imputed_vals = tf.gather(self.impute_scalars[0, :],
                                 self.missing_ind[:, 1])

        # Fill the imputed values into the data tensor of zeros
        shape = tf.cast(tf.shape(data_zeroed_missing), dtype=tf.int64)
        missing_imputed = tf.scatter_nd(self.missing_ind, imputed_vals, shape)

        X_with_impute = data_zeroed_missing + missing_imputed

        return X_with_impute


class LearnedNormalImpute(ImputeOp):
    r"""Impute the missing values with draws from learned normal distributions.

    Takes two layers, one the returns a data tensor and the other returns a
    mask layer. This creates a layer that will learn marginal Gaussian
    parameters per column, and infill missing values using draws from these
    Gaussians.

    Parameters
    ----------
    datalayer : callable
        A layer that returns a data tensor. Must be an InputLayer.
    masklayer : callable
        A layer that returns a boolean mask tensor where True values are
        masked. Must be an InputLayer.
    """

    def __init__(self, datalayer, masklayer):
        r"""Construct and instance of a VarScalarImpute operation."""
        super().__init__(datalayer, masklayer)

    def _initialise_variables(self, X):
        """Initialise the impute variables."""
        datadim = int(X.shape[2])
        impute_means = tf.Variable(
            tf.random_normal(shape=(1, datadim), seed=next(seedgen)),
            name="impute_scalars"
        )
        impute_variances = tf.Variable(
            tf.random_normal(shape=(1, datadim), seed=next(seedgen)),
            name="impute_scalars"
        )
        self.normal = Normal(impute_means, pos(impute_variances))

    def _impute2D(self, X_2D):
        r"""Impute a rank 2 tensor with draws from normal distributions.

        Parameters
        ----------
        X_2D : Tensor
            a rank 2 Tensor with missing data

        Returns
        -------
        X_imputed : Tensor
            a rank 2 Tensor with imputed data

        """
        # Fill zeros in for missing data initially
        data_zeroed_missing = X_2D * self.real_val_mask

        # Divide column totals by the number of non-nan values
        col_draws = tf.transpose(self.normal.sample())
        # Make an vector of the impute values for each missing point
        imputed_vals = tf.gather(col_draws, self.missing_ind[:, 1])[:, 0]

        # Fill the imputed values into the data tensor of zeros
        shape = tf.cast(tf.shape(data_zeroed_missing), dtype=tf.int64)
        missing_imputed = tf.scatter_nd(self.missing_ind, imputed_vals, shape)

        X_with_impute = data_zeroed_missing + missing_imputed

        return X_with_impute



"""Network layers and utilities."""
import numpy as np
import tensorflow as tf

from aboleth.kernels import RBF, RBFVariational
from aboleth.random import seedgen
from aboleth.distributions import (norm_prior, norm_posterior, gaus_posterior,
                                   kl_qp)
from aboleth.baselayers import Layer, MultiLayer


#
# Layer type base classes
#

class InputLayer(MultiLayer):
    r"""Create an input layer.

    This layer defines input kwargs so that a user may easily provide the right
    inputs to a complex set of layers. It takes a 2D tensor of shape (N, D).
    If n_samples is specified, the input is tiled along a new first axis
    creating a (n_samples, N, D) tensor for propogating samples through a
    variational deep net.

    Parameters
    ----------
    name : string
        The name of the input. Used as the agument for input into the net.
    n_samples : int > 0
        The number of samples.

    """

    def __init__(self, name, n_samples=None):
        """Construct an instance of InputLayer."""
        self.name = name
        self.n_samples = n_samples

    def _build(self, **kwargs):
        """Build the tiling input layer."""
        X = kwargs[self.name]
        if self.n_samples is not None:
            # (n_samples, N, D)
            Xs = tf.tile(tf.expand_dims(X, 0), [self.n_samples, 1, 1])
        else:
            Xs = tf.convert_to_tensor(X)
        return Xs, 0.0


class SampleLayer(Layer):
    r"""Sample Layer base class.

    This is the base class for layers that build upon stochastic (variational)
    nets. These expect *rank >= 3* input Tensors, where the first dimension
    indexes the random samples of the stochastic net.
    """

    def __call__(self, X):
        r"""Construct the subgraph for this layer.

        Parameters
        ----------
        X : Tensor
            the input to this layer

        Returns
        -------
        Net : Tensor
            the output of this layer
        KL : float, Tensor
            the regularizer/Kullback Leibler 'cost' of the parameters in this
            layer.

        """
        rank = len(X.shape)
        assert rank > 2
        Net, KL = self._build(X)
        return Net, KL

    @staticmethod
    def _get_X_dims(X):
        r"""Get the dimensions of the rank >= 3 input tensor, X."""
        n_samples, _, *input_shape = X.shape.as_list()
        return n_samples, input_shape


class SampleLayer3(SampleLayer):
    r"""Special case of SampleLayer restricted to *rank == 3* input Tensors."""

    def __call__(self, X):
        r"""Construct the subgraph for this layer.

        Parameters
        ----------
        X : Tensor
            the input to this layer

        Returns
        -------
        Net : Tensor
            the output of this layer
        KL : float, Tensor
            the regularizer/Kullback Leibler 'cost' of the parameters in this
            layer.

        """
        rank = len(X.shape)
        assert rank == 3
        Net, KL = super(SampleLayer3, self).__call__(X)
        return Net, KL

    @staticmethod
    def _get_X_dims(X):
        """Get the dimensions of the rank 3 input tensor, X."""
        n_samples, _, input_dim = X.shape.as_list()
        return n_samples, input_dim


#
# Activation Layers
#

class Activation(Layer):
    """Activation function layer.

    Parameters
    ----------
    h : callable
        the *element-wise* activation function.

    """

    def __init__(self, h=lambda X: X):
        """Create an instance of an Activation layer."""
        self.h = h

    def _build(self, X):
        """Build the graph of this layer."""
        Net = self.h(X)
        KL = 0.
        return Net, KL


class DropOut(Layer):
    r"""Dropout layer, Bernoulli probability of not setting an input to zero.

    This is just a thin wrapper around `tf.dropout
    <https://www.tensorflow.org/api_docs/python/tf/nn/dropout>`_

    Parameters
    ----------
    keep_prob : float, Tensor
        the probability of keeping an input. See `tf.dropout
        <https://www.tensorflow.org/api_docs/python/tf/nn/dropout>`_.

    """

    def __init__(self, keep_prob):
        """Create an instance of a Dropout layer."""
        self.keep_prob = keep_prob

    def _build(self, X):
        """Build the graph of this layer."""
        noise_shape = None  # equivalent to different samples from posterior
        Net = tf.nn.dropout(X, self.keep_prob, noise_shape, seed=next(seedgen))
        KL = 0.
        return Net, KL


class MaxPool2D(Layer):
    r"""Max pooling layer for 2D inputs (e.g. images).

    This is just a thin wrapper around `tf.nn.max_pool
    <https://www.tensorflow.org/api_docs/python/tf/nn/max_pool>`_

    Parameters
    ----------
    pool_size : tuple or list of 2 ints
        width and height of the pooling window.
    strides : tuple or list of 2 ints
        the strides of the pooling operation along the height and width.
    padding : str
        One of 'SAME' or 'VALID'. Defaults to 'SAME'. The type of padding

    """

    def __init__(self, pool_size, strides, padding='SAME'):
        """Initialize instance of a MaxPool2D layer."""
        self.ksize = [1] + list(pool_size) + [1]
        self.strides = [1] + list(strides) + [1]
        self.padding = padding

    def _build(self, X):
        """Build the graph of this layer."""
        Net = tf.map_fn(lambda inputs: tf.nn.max_pool(inputs,
                                                      ksize=self.ksize,
                                                      strides=self.strides,
                                                      padding=self.padding), X)
        KL = 0.
        return Net, KL


class Reshape(Layer):
    """Reshape layer.

    Reshape and output an tensor to a specified shape.

    Parameters
    ----------
    targe_shape : tuple of ints
        Does not include the samples or batch axes.

    """

    def __init__(self, target_shape):
        """Initialize instance of a Reshape layer."""
        self.target_shape = target_shape

    def _build(self, X):
        """Build the graph of this layer."""
        new_shape = X.shape[:2].concatenate(tf.TensorShape(self.target_shape))
        Net = tf.reshape(X, new_shape)
        KL = 0.
        return Net, KL


#
# Kernel Approximation Layers
#

class RandomFourier(SampleLayer3):
    r"""Random Fourier feature (RFF) kernel approximation layer.

    NOTE: This should be followed by a dense layer to properly implement a
        kernel approximation.

    Parameters
    ----------
    n_features : int
        the number of unique random features, the actual output dimension of
        this layer will be ``2 * n_features``.
    kernel : kernels.ShiftInvariant
        the kernel object that yeilds the random samples from the fourier
        spectrum of a particular kernel to approximate. See the :ref:`kernels`
        module.

    """

    def __init__(self, n_features, kernel):
        """Construct and instance of a RandomFourier object."""
        self.n_features = n_features
        self.kernel = kernel

    def _build(self, X):
        """Build the graph of this layer."""
        # Random weights
        n_samples, input_dim = self._get_X_dims(X)
        P, KL = self.kernel.weights(input_dim, self.n_features)
        Ps = tf.tile(tf.expand_dims(P, 0), [n_samples, 1, 1])

        # Random features
        XP = tf.matmul(X, Ps)
        Net = self._transformation(XP)
        return Net, KL

    def _transformation(self, XP):
        """Build the kernel feature space transformation."""
        real = tf.cos(XP)
        imag = tf.sin(XP)
        Net = tf.concat([real, imag], axis=-1) / np.sqrt(self.n_features)
        return Net


class RandomArcCosine(RandomFourier):
    r"""Random arc-cosine kernel layer.

    NOTE: This should be followed by a dense layer to properly implement a
        kernel approximation.

    Parameters
    ----------
    n_features : int
        the number of unique random features, the actual output dimension of
        this layer will be ``2 * n_features``.
    lenscale : float, ndarray, Tensor
        the lenght scales of the ar-cosine kernel, this can be a scalar for
        an isotropic kernel, or a vector for an automatic relevance detection
        (ARD) kernel.
    p : int
        The order of the arc-cosine kernel, this must be an integer greater
        than, or eual to zero. 0 will lead to sigmoid-like kernels, 1 will lead
        to relu-like kernels, 2 quadratic-relu kernels etc.
    variational : bool
        use variational features instead of random features, (i.e. VAR-FIXED in
        [2]).
    lenscale_posterior : float, ndarray, optional
        the *initial* value for the posterior length scale. This is only used
        if ``variational==True``. This can be a scalar or vector (different
        initial value per input dimension). If this is left as None, it will be
        set to ``sqrt(1 / input_dim)`` (this is similar to the 'auto' setting
        for a scikit learn SVM with a RBF kernel).

    See Also
    --------
    [1] Cho, Youngmin, and Lawrence K. Saul.
        "Analysis and extension of arc-cosine kernels for large margin
        classification." arXiv preprint arXiv:1112.3712 (2011).
    [2] Cutajar, K. Bonilla, E. Michiardi, P. Filippone, M.
        Random Feature Expansions for Deep Gaussian Processes. In ICML, 2017.

    """

    def __init__(self, n_features, lenscale=1.0, p=1, variational=False,
                 lenscale_posterior=None):
        """Create an instance of an arc cosine kernel layer."""
        # Setup random weights
        if variational:
            kern = RBFVariational(lenscale=lenscale,
                                  lenscale_posterior=lenscale_posterior)
        else:
            kern = RBF(lenscale=lenscale)
        super().__init__(n_features=n_features, kernel=kern)

        # Kernel order
        assert isinstance(p, int) and p >= 0
        if p == 0:
            self.pfunc = tf.sign
        elif p == 1:
            self.pfunc = lambda x: x
        else:
            self.pfunc = lambda x: tf.pow(x, p)

    def _transformation(self, XP):
        """Build the kernel feature space transformation."""
        Net = np.sqrt(2. / self.n_features) * tf.nn.relu(self.pfunc(XP))
        return Net


#
# Weight layers
#

class DenseVariational(SampleLayer3):
    r"""Dense (fully connected) linear layer, with variational inference.

    Parameters
    ----------
    output_dim : int
        the dimension of the output of this layer
    var : float
        the initial value of the weight prior variance, which defaults to
        :math:`\mathbf{W} \sim \mathcal{N}(\mathbf{0}, \text{var} \mathbf{I})`,
        this is optimized (a la maximum likelihood type II).
    full : bool
        If true, use a full covariance Gaussian posterior for *each* of the
        output weight columns, otherwise use an independent (diagonal) Normal
        posterior.
    use_bias : bool
        If true, also learn a bias weight, e.g. a constant offset weight.
    prior_W : distributions.Normal, distributions.Gaussian, optional
        This is the prior distribution object to use on the layer weights. It
        must have parameters compatible with (input_dim, output_dim) shaped
        weights. This ignores the ``var`` parameter.
    prior_b : distributions.Normal, distributions.Gaussian, optional
        This is the prior distribution object to use on the layer intercept. It
        must have parameters compatible with (output_dim,) shaped weights.
        This ignores the ``var`` and ``use_bias`` parameters.
    post_W : distributions.Normal, distributions.Gaussian, optional
        It must have parameters compatible with (input_dim, output_dim) shaped
        weights. This ignores the ``full`` parameter. See also
        ``distributions.gaus_posterior``.
    post_b : distributions.Normal, distributions.Gaussian, optional
        This is the posterior distribution object to use on the layer
        intercept. It must have parameters compatible with (output_dim,) shaped
        weights. This ignores the ``use_bias`` parameters.  See also
        ``distributions.norm_posterior``.

    """

    def __init__(self, output_dim, var=1., full=False, use_bias=True,
                 prior_W=None, prior_b=None, post_W=None, post_b=None):
        """Create and instance of a variational dense layer."""
        self.output_dim = output_dim
        self.var = var
        self.full = full
        self.use_bias = use_bias
        self.pW = prior_W
        self.pb = prior_b
        self.qW = post_W
        self.qb = post_b

    def _build(self, X):
        """Build the graph of this layer."""
        n_samples, input_dim = self._get_X_dims(X)
        W_shape, b_shape = self._weight_shapes(input_dim)

        # Layer weights
        self.pW = self._make_prior(self.pW, W_shape)
        self.qW = self._make_posterior(self.qW, W_shape)

        # Regularizers
        KL = kl_qp(self.qW, self.pW)

        # Linear layer
        Wsamples = self._sample_W(self.qW, n_samples)
        Net = tf.matmul(X, Wsamples)

        # Optional bias
        if self.use_bias is True or self.prior_b or self.post_b:
            # Layer intercepts
            self.pb = self._make_prior(self.pb, b_shape)
            self.qb = self._make_posterior(self.qb, b_shape)

            # Regularizers
            KL += kl_qp(self.qb, self.pb)

            # Linear layer
            bsamples = tf.expand_dims(self._sample_W(self.qb, n_samples), 1)
            Net += bsamples

        return Net, KL

    def _make_prior(self, prior_W, weight_shape):
        """Check/make prior."""
        if prior_W is None:
            prior_W = norm_prior(dim=weight_shape, var=self.var)

        assert _is_dim(prior_W.mu, weight_shape), \
            "Prior inconsistent dimension!"

        return prior_W

    def _make_posterior(self, post_W, weight_shape):
        """Check/make posterior."""
        if post_W is None:
            # We don't want a full-covariance on an intercept, check input_dim
            if self.full and len(weight_shape) > 1:
                post_W = gaus_posterior(dim=weight_shape, var0=self.var)
            else:
                post_W = norm_posterior(dim=weight_shape, var0=self.var)

        assert _is_dim(post_W.mu, weight_shape), \
            "Posterior inconsistent dimension!"

        return post_W

    def _weight_shapes(self, input_dim):
        """Generate weight and bias weight shape tuples."""
        weight_shape = (input_dim, self.output_dim)
        bias_shape = (self.output_dim,)

        return weight_shape, bias_shape

    @staticmethod
    def _sample_W(dist, n_samples):
        samples = tf.stack([dist.sample() for _ in range(n_samples)])
        return samples


class EmbedVariational(DenseVariational):
    r"""Dense (fully connected) embedding layer, with variational inference.

    This layer works directly on shape (N, 1) inputs of category *indices*
    rather than one-hot representations, for efficiency.

    Parameters
    ----------
    output_dim : int
        the dimension of the output (embedding) of this layer
    n_categories : int
        the number of categories in the input variable
    var : float
        the initial value of the weight prior variance, which defaults to
        :math:`\mathbf{W} \sim \mathcal{N}(\mathbf{0}, \text{var} \mathbf{I})`,
        this is optimized (a la maximum likelihood type II).
    full : bool
        If true, use a full covariance Gaussian posterior for *each* of the
        output weight columns, otherwise use an independent (diagonal) Normal
        posterior.
    prior_W : distributions.Normal, distributions.Gaussian, optional
        This is the prior distribution object to use on the layer weights. It
        must have parameters compatible with (input_dim, output_dim) shaped
        weights. This ignores the ``var`` parameter.
    post_W : distributions.Normal, distributions.Gaussian, optional
        This is the posterior distribution object to use on the layer weights.
        It must have parameters compatible with (input_dim, output_dim) shaped
        weights. This ignores the ``full`` parameter. See also
        ``distributions.gaus_posterior``.

    """

    def __init__(self, output_dim, n_categories, var=1., full=False,
                 prior_W=None, post_W=None):
        """Create and instance of a variational dense embedding layer."""
        assert n_categories >= 2, "Need 2 or more categories for embedding!"
        self.output_dim = output_dim
        self.n_categories = n_categories
        self.var = var
        self.full = full
        self.pW = prior_W
        self.qW = post_W

    def _build(self, X):
        """Build the graph of this layer."""
        n_samples, input_dim = self._get_X_dims(X)
        W_shape, _ = self._weight_shapes(self.n_categories)

        assert input_dim == 1, "X must be a *column* of indices!"

        # Layer weights
        self.pW = self._make_prior(self.pW, W_shape)
        self.qW = self._make_posterior(self.qW, W_shape)

        # Index into the relevant weights rather than using sparse matmul
        Wsamples = self._sample_W(self.qW, n_samples)
        Net = tf.gather(Wsamples, X[0, :, 0], axis=1)

        # Regularizers
        KL = kl_qp(self.qW, self.pW)

        return Net, KL


class DenseMAP(SampleLayer):
    r"""Dense (fully connected) linear layer, with MAP inference.

    Parameters
    ----------
    output_dim : int
        the dimension of the output of this layer
    l1_reg : float
        the value of the l1 weight regularizer,
        :math:`\text{l1_reg} \times \|\mathbf{W}\|_1`
    l2_reg : float
        the value of the l2 weight regularizer,
        :math:`\frac{1}{2} \text{l2_reg} \times \|\mathbf{W}\|^2_2`
    use_bias : bool
        If true, also learn a bias weight, e.g. a constant offset weight.

    """

    def __init__(self, output_dim, l1_reg=1., l2_reg=1., use_bias=True):
        """Create and instance of a dense layer with MAP regularizers."""
        self.output_dim = output_dim
        self.l1 = l1_reg
        self.l2 = l2_reg
        self.use_bias = use_bias

    def _build(self, X):
        """Build the graph of this layer."""
        n_samples, input_shape = self._get_X_dims(X)

        Wdim = tuple(input_shape) + (self.output_dim,)

        W = tf.Variable(tf.random_normal(shape=Wdim, seed=next(seedgen)),
                        name="W_map")

        # We don't want to copy tf.Variable W so map over X
        Net = tf.map_fn(lambda x: tf.matmul(x, W), X)

        # Regularizers
        penalty = self.l2 * tf.nn.l2_loss(W) + self.l1 * _l1_loss(W)

        # Optional Bias
        if self.use_bias is True:
            b = tf.Variable(tf.random_normal(shape=(1, self.output_dim),
                                             seed=next(seedgen)), name="b_map")
            Net += b
            penalty += self.l2 * tf.nn.l2_loss(b) + self.l1 * _l1_loss(b)

        return Net, penalty


#
# Private module stuff
#

def _l1_loss(X):
    r"""Calculate the L1 loss of X, :math:`\|X\|_1`."""
    l1 = tf.reduce_sum(tf.abs(X))
    return l1


def _is_dim(X, dims):
    r"""Check if ``X``'s dimension is the same as the tuple ``dims``."""
    shape = tuple([int(d) for d in X.get_shape()])
    return shape == dims



"""Model parameter distributions."""
import numpy as np
import tensorflow as tf
from multipledispatch import dispatch

from aboleth.util import pos
from aboleth.random import seedgen


#
# Generic prior and posterior classes
#

class ParameterDistribution:
    """Abstract base class for parameter distribution objects."""

    def __init__(self):
        """Construct a ParameterDistibution object."""
        raise NotImplementedError('Abstract base class only.')

    def sample(self):
        """Draw a random sample from the distribution."""
        raise NotImplementedError('Abstract base class only.')


class Normal(ParameterDistribution):
    """
    Normal (IID) prior/posterior.

    Parameters
    ----------
    mu : Tensor
        mean, shape (d_in, d_out)
    var : Tensor
        variance, shape (d_in, d_out)

    """

    def __init__(self, mu=0., var=1.):
        """Construct a Normal distribution object."""
        self.mu = mu
        self.var = var
        self.sigma = tf.sqrt(var)
        self.d = mu.shape

    def sample(self, e=None):
        """Draw a random sample from this object.

        Parameters
        ----------
        e : ndarray, Tensor, optional
            the random standard-Normal samples to transform to yeild samples
            from this distrubution. These must be of shape (d_in, ...). If
            this is none, these are generated in this method.

        Returns
        -------
        x : Tensor
            a sample of shape (d_in, d_out), or ``e.shape`` if provided

        """
        # Reparameterisation trick
        if e is None:
            e = tf.random_normal(self.d, seed=next(seedgen))
        x = self.mu + e * self.sigma

        return x


class Gaussian(ParameterDistribution):
    """
    Gaussian prior/posterior.

    Parameters
    ----------
    mu : Tensor
        mean, shape (d_in, d_out)
    L : Tensor
        Cholesky of the covariance matrix, shape (d_out, d_in, d_in)
    eps : ndarray, Tensor, optional
        random draw from a unit normal if you want to "fix" the sampling, this
        should be of shape (d_in, d_out). If this is ``None`` then a new random
        draw is used for every call to sample().

    """

    def __init__(self, mu, L):
        """Construct a Normal distribution object."""
        self.mu = mu
        self.L = L  # O x I x I
        self.d = mu.shape

    def sample(self, e=None):
        """Draw a random sample from this object.

        Parameters
        ----------
        e : ndarray, Tensor, optional
            the random standard-Normal samples to transform to yeild samples
            from this distrubution. These must be of shape (d_in, ...). If
            this is none, these are generated in this method.

        Returns
        -------
        x : Tensor
            a sample of shape (d_in, d_out), or ``e.shape`` if provided

        """
        # Reparameterisation trick
        mu = self.transform_w(self.mu)
        if e is None:
            e = tf.random_normal(mu.shape, seed=next(seedgen))
        else:
            e = self.transform_w(e)
        x = self.itransform_w(mu + tf.matmul(self.L, e))

        return x

    @staticmethod
    def transform_w(w):
        """Transform a weight matrix, (d_in, d_out) -> (d_out, d_in, 1)."""
        wt = tf.expand_dims(tf.transpose(w), 2)  # O x I x 1
        return wt

    @staticmethod
    def itransform_w(wt):
        """Un-transform a weight matrix, (d_out, d_in, 1) -> (d_in, d_out)."""
        w = tf.transpose(wt[:, :, 0])
        return w


#
# Streamlined interfaces for initialising the priors and posteriors
#

def norm_prior(dim, var):
    """Initialise a prior (zero mean, diagonal) Normal distribution.

    Parameters
    ----------
    dim : tuple or list
        the dimension of this distribution.
    var : float
        the prior variance of this distribution.

    Returns
    -------
    P : Normal
        the initialised prior Normal object.

    Note
    ----
    This will make a tf.Variable on the variance of the prior that is
    initialised with ``var``.

    """
    mu = tf.zeros(dim)
    var = pos(tf.Variable(var, name="W_mu_p"))
    P = Normal(mu, var)
    return P


def norm_posterior(dim, var0):
    """Initialise a posterior (diagonal) Normal distribution.

    Parameters
    ----------
    dim : tuple or list
        the dimension of this distribution.
    var0 : float
        the initial (unoptimized) variance of this distribution.

    Returns
    -------
    Q : Normal
        the initialised posterior Normal object.

    Note
    ----
    This will make tf.Variables on the randomly initialised mean and variance
    of the posterior. The initialisation of the mean is from a Normal with zero
    mean, and ``var0`` variance, and the initialisation of the variance is from
    a gamma distribution with an alpha of ``var0`` and a beta of 1.

    """
    mu_0 = tf.random_normal(dim, stddev=tf.sqrt(var0), seed=next(seedgen))
    mu = tf.Variable(mu_0, name="W_mu_q")

    var_0 = tf.random_gamma(alpha=var0, shape=dim, seed=next(seedgen))
    var = pos(tf.Variable(var_0, name="W_var_q"))

    Q = Normal(mu, var)
    return Q


def gaus_posterior(dim, var0):
    """Initialise a posterior Gaussian distribution with a diagonal covariance.

    Even though this is initialised with a diagonal covariance, a full
    covariance will be learned, using a lower triangular Cholesky
    parameterisation.

    Parameters
    ----------
    dim : tuple or list
        the dimension of this distribution.
    var0 : float
        the initial (unoptimized) diagonal variance of this distribution.

    Returns
    -------
    Q : Gaussian
        the initialised posterior Gaussian object.

    Note
    ----
    This will make tf.Variables on the randomly initialised mean and covariance
    of the posterior. The initialisation of the mean is from a Normal with zero
    mean, and ``var0`` variance, and the initialisation of the variance is from
    a gamma distribution with an alpha of ``var0`` and a beta of 1.

    """
    I, O = dim
    sig0 = np.sqrt(var0)

    # Optimize only values in lower triangular
    u, v = np.tril_indices(I)
    indices = (u * I + v)[:, np.newaxis]
    l0 = np.tile(np.eye(I), [O, 1, 1])[:, u, v].T
    l0 = l0 * tf.random_gamma(alpha=var0, shape=l0.shape, seed=next(seedgen))
    l = tf.Variable(l0, name="W_cov_q")
    Lt = tf.transpose(tf.scatter_nd(indices, l, shape=(I * I, O)))
    L = tf.reshape(Lt, (O, I, I))

    mu_0 = tf.random_normal((I, O), stddev=sig0, seed=next(seedgen))
    mu = tf.Variable(mu_0, name="W_mu_q")
    Q = Gaussian(mu, L)
    return Q


#
# KL divergence calculations
#


@dispatch(Normal, Normal)
def kl_qp(q, p):
    """Normal-Normal Kullback Leibler divergence calculation.

    Parameters
    ----------
    q : Normal
        the approximating 'q' distribution.
    p : Normal
        the prior 'p' distribution.

    Returns
    -------
    KL : Tensor
        the result of KL[q||p].

    """
    KL = 0.5 * (tf.log(p.var) - tf.log(q.var) + q.var / p.var - 1. +
                (q.mu - p.mu)**2 / p.var)
    KL = tf.reduce_sum(KL)
    return KL


@dispatch(Gaussian, Normal)  # noqa
def kl_qp(q, p):
    """Gaussian-Normal Kullback Leibler divergence calculation.

    Parameters
    ----------
    q : Gaussian
        the approximating 'q' distribution.
    p : Normal
        the prior 'p' distribution.

    Returns
    -------
    KL : Tensor
        the result of KL[q||p].

    """
    D, n = tf.to_float(q.d[0]), tf.to_float(q.d[1])
    tr = tf.reduce_sum(q.L * q.L) / p.var
    dist = tf.reduce_sum((p.mu - q.mu)**2) / p.var
    logdet = n * D * tf.log(p.var) - _chollogdet(q.L)
    KL = 0.5 * (tr + dist + logdet - n * D)
    return KL


@dispatch(Gaussian, Gaussian)  # noqa
def kl_qp(q, p):
    """Gaussian-Gaussian Kullback Leibler divergence calculation.

    Parameters
    ----------
    q : Gaussian
        the approximating 'q' distribution.
    p : Gaussian
        the prior 'p' distribution.

    Returns
    -------
    KL : Tensor
        the result of KL[q||p].

    """
    D, n = tf.to_float(q.d[0]), tf.to_float(q.d[1])
    qCipC = tf.cholesky_solve(p.L, tf.matmul(q.L, q.L, transpose_b=True))
    tr = tf.reduce_sum(tf.trace(qCipC))
    md = q.transform_w(p.mu - q.mu)
    dist = tf.reduce_sum(md * tf.cholesky_solve(p.L, md))
    logdet = _chollogdet(p.L) - _chollogdet(q.L)
    KL = 0.5 * (tr + dist + logdet - n * D)
    return KL


#
# Private module stuff
#

def _chollogdet(L):
    """Log det of a cholesky, where L is (..., D, D)."""
    l = tf.maximum(tf.matrix_diag_part(L), 1e-15)  # Make sure we don't go to 0
    logdet = 2. * tf.reduce_sum(tf.log(l))
    return logdet



"""Test the baselayers module."""
import tensorflow as tf
import aboleth as ab

from .util import StringLayer, StringMultiLayer


def test_stack2():
    """Test base implementation of stack."""
    def f(X):
        return "f({})".format(X), 10.0

    def g(X):
        return "g({})".format(X), 20.0

    h = ab.baselayers._stack2(f, g)

    tc = tf.test.TestCase()
    with tc.test_session():
        phi, loss = h(X="x")
        assert phi == "g(f(x))"
        assert loss.eval() == 30.0


def test_stack2_multi():
    """Test base implementation of stack."""
    def f(X, Y):
        return "f({}, {})".format(X, Y), 10.0

    def g(X):
        return "g({})".format(X), 20.0

    h = ab.baselayers._stack2(f, g)

    tc = tf.test.TestCase()
    with tc.test_session():
        phi, loss = h(X="x", Y="y")
        assert phi == "g(f(x, y))"
        assert loss.eval() == 30.0


def test_stack_layer():
    """Test implementation of stack."""
    f = StringLayer("f", 10.0)
    g = StringLayer("g", 20.0)
    h = StringLayer("h", 30.0)
    r = ab.stack(f, g, h)

    tc = tf.test.TestCase()
    with tc.test_session():
        phi, loss = r("x")
        assert phi == "h(g(f(x)))"
        assert loss.eval() == 60.0


def test_stack_multilayer():
    """Test implementation of stack."""
    f = StringMultiLayer("f", ["x", "y"], 10.0)
    g = StringLayer("g", 20.0)
    h = StringLayer("h", 30.0)
    r = ab.stack(f, g, h)
    tc = tf.test.TestCase()
    with tc.test_session():
        phi, loss = r(x="x", y="y")
        assert phi == "h(g(f(x,y)))"
        assert loss.eval() == 60.0



from aboleth.baselayers import Layer, MultiLayer


class StringLayer(Layer):

    def __init__(self, name='f', kl=0.0):
        self.name = name
        self.kl = kl

    def _build(self, X):
        result = "{}({})".format(self.name, X)
        return result, self.kl


class StringMultiLayer(MultiLayer):

    def __init__(self, name='f', args=None, kl=0.0):
        args = ['x'] if not args else args
        self.name = name
        self.args = args
        self.kl = kl

    def _build(self, **kwargs):
        l = [kwargs[i] for i in sorted(kwargs)]
        call_sig = ",".join(l)
        result = "{}({})".format(self.name, call_sig)
        return result, self.kl



"""Package helper utilities."""
import tensorflow as tf
import numpy as np

from aboleth.random import endless_permutations


def pos(X, minval=1e-15):
    r"""Constrain a ``tf.Variable`` to be positive only.

    At the moment this is implemented as:

        :math:`\max(|\mathbf{X}|, \text{minval})`

    This is fast and does not result in vanishing gradients, but will lead to
    non-smooth gradients and more local minima. In practice we haven't noticed
    this being a problem.

    Parameters
    ----------
    X : Tensor
        any Tensor in which all elements will be made positive.
    minval : float
        the minimum "positive" value the resulting tensor will have.

    Returns
    -------
    X : Tensor
        a tensor the same shape as the input ``X`` but positively constrained.

    Examples
    --------
    >>> X = tf.constant(np.array([1.0, -1.0, 0.0]))
    >>> Xp = pos(X)
    >>> with tf.Session():
    ...     xp = Xp.eval()
    >>> xp
    array([  1.00000000e+00,   1.00000000e+00,   1.00000000e-15])

    """
    # Other alternatives could be:
    # Xp = tf.exp(X)  # Medium speed, but gradients tend to explode
    # Xp = tf.nn.softplus(X)  # Slow but well behaved!
    Xp = tf.maximum(tf.abs(X), minval)  # Faster, but more local optima
    return Xp


def batch(feed_dict, batch_size, n_iter=10000, N_=None):
    r"""Create random batches for Stochastic gradients.

    Feed dict data generator for SGD that will yeild random batches for a
    a defined number of iterations, which can be infinite. This generator makes
    consecutive passes through the data, drawing without replacement on each
    pass.

    Parameters
    ----------
    feed_dict : dict of ndarrays
        The data with ``{tf.placeholder: data}`` entries. This assumes all
        items have the *same* length!
    batch_size : int
        number of data points in each batch.
    n_iter : int, optional
        The number of iterations
    N_ : tf.placeholder (int), optional
        Place holder for the size of the dataset. This will be fed to an
        algorithm.

    Yields
    ------
    dict:
        with each element an array length ``batch_size``, i.e. a subset of
        data, and an element for ``N_``. Use this as your feed-dict when
        evaluating a loss, training, etc.

    """
    N = __data_len(feed_dict)
    perms = endless_permutations(N)

    i = 0
    while i < n_iter:
        i += 1
        ind = np.array([next(perms) for _ in range(batch_size)])
        batch_dict = {k: v[ind] for k, v in feed_dict.items()}
        if N_ is not None:
            batch_dict[N_] = N
        yield batch_dict


def batch_prediction(feed_dict, batch_size):
    r"""Split the data in a feed_dict into contiguous batches for prediction.

    Parameters
    ----------
    feed_dict : dict of ndarrays
        The data with ``{tf.placeholder: data}`` entries. This assumes all
        items have the *same* length!
    batch_size : int
        number of data points in each batch.

    Yields
    ------
    ndarray :
        an array of shape approximately (``batch_size``,) of indices into the
        original data for the current batch
    dict :
        with each element an array length ``batch_size``, i.e. a subset of
        data. Use this as your feed-dict when evaluating a model, prediction,
        etc.

    Note
    ----
    The exact size of the batch may not be ``batch_size``, but the nearest size
    that splits the size of the data most evenly.

    """
    N = __data_len(feed_dict)
    n_batches = max(np.round(N / batch_size), 1)
    batch_inds = np.array_split(np.arange(N, dtype=int), n_batches)

    for ind in batch_inds:
        batch_dict = {k: v[ind] for k, v in feed_dict.items()}
        yield ind, batch_dict


def predict_samples(predictor, feed_dict, n_groups=1, session=None):
    r"""Help to get samples from a predictor.

    Parameters
    ----------
    predictor : Tensor
        a tensor that outputs a shape (n_samples, N, tasks) where
        ``n_samples`` are the random samples from the predictor (e.g. the
        output of ``Net``), ``N`` is the size of the query dataset, and
        ``tasks`` the number of prediction tasks.
    feed_dict : dict
        The data with ``{tf.placeholder: data}`` entries.
    n_groups : int
        The number of times to evaluate the ``predictor`` and concatenate the
        samples.
    session : Session
        the session to be used to evaluate the predictor.

    Returns
    -------
    pred : ndarray
        prediction samples of shape (n_samples * n_groups, N, tasks).

    Note
    ----
    This has to be called in an *active* tensorflow session!

    """
    pred = [predictor.eval(feed_dict=feed_dict, session=session)
            for _ in range(n_groups)]
    pred = np.concatenate(pred, axis=0)
    return pred


def predict_expected(predictor, feed_dict, n_groups=1, session=None):
    r"""Help to get the expected value from a predictor.

    Parameters
    ----------
    predictor : Tensor
        a tensor that outputs a shape (n_samples, N, tasks) where
        ``n_samples`` are the random samples from the predictor (e.g. the
        output of ``Net``), ``N`` is the size of the query dataset, and
        ``tasks`` the number of prediction tasks.
    feed_dict : dict
        The data with ``{tf.placeholder: data}`` entries.
    n_groups : int
        The number of times to evaluate the ``predictor`` and concatenate the
        samples.
    session : Session
        the session to be used to evaluate the predictor.

    Returns
    -------
    pred : ndarray
        expected value of the prediction with shape (N, tasks). ``n_samples *
        n_groups`` samples go into evaluating this expectation.

    Note
    ----
    This has to be called in an *active* tensorflow session!

    """
    pred = 0
    for _ in range(n_groups):
        pred += predictor.eval(feed_dict=feed_dict, session=session) / n_groups

    pred = pred.mean(axis=0)
    return pred


def __data_len(feed_dict):
    N = feed_dict[list(feed_dict.keys())[0]].shape[0]
    return N



"""Test the likelihood module."""
import pytest
import numpy as np
import tensorflow as tf
import scipy.stats as ss
from scipy.special import expit

from aboleth.likelihoods import Normal, Bernoulli, Binomial, Categorical


@pytest.mark.parametrize('likelihood', [
    (Normal(variance=1),
     ss.norm.rvs,
     lambda x, f: ss.norm.logpdf(x, loc=f)),

    (Binomial(n=10.),
     lambda f, size: ss.binom.rvs(n=10, p=f, size=size),
     lambda x, f: ss.binom.logpmf(x, n=10, p=f)),

    (Bernoulli(),
     ss.bernoulli.rvs,
     lambda x, f: ss.bernoulli.logpmf(x, p=f)),

])
def test_log_likelihoods(likelihood, random):

    alike, rvs, logprob = likelihood

    f = expit(random.randn(100).astype(np.float32))
    x = rvs(f, size=100).astype(np.float32)

    tc = tf.test.TestCase()
    with tc.test_session():
        assert np.allclose(logprob(x, f), alike(x, f).eval())


@pytest.mark.parametrize('likelihood', [
    (Categorical(),
     lambda f, size: ss.multinomial.rvs(n=1, p=f, size=size),
     lambda x, f: ss.multinomial.logpmf(x, n=1, p=f)),
])
def test_log_likelihoods_multitask(likelihood, random):

    alike, rvs, logprob = likelihood

    f = expit(random.randn(100, 5))
    f /= np.sum(f, axis=-1).reshape(-1, 1)  # normalize
    f = f.astype(np.float32)

    # apply_along_axis as scipy rvs doesn't support broadcasting
    x = np.apply_along_axis(lambda f: rvs(f, size=1), arr=f, axis=1)
    x = x.astype(np.float32)

    tc = tf.test.TestCase()
    with tc.test_session():
        assert np.allclose(logprob(x, f), alike(x, f).eval())



"""Target/Output likelihoods."""
import numpy as np
import tensorflow as tf

from aboleth.util import pos


class Likelihood:
    """Abstract base class for likelihood objects."""

    def __call__(self, y, f):
        """Build the log likelihood.

        See: _loglike.

        """
        ll = self._loglike(y, f)
        return ll

    def _loglike(self, y, f):
        """Build the log likelihood.

        Parameters
        ----------
        y : Tensor
            the target variable of shape (N, tasks)
        f : Tensor
            the latent function output from the network of shape (N, tasks)

        """
        raise NotImplementedError('Abstract base class only.')


class Normal(Likelihood):
    """Normal log-likelihood.

    Parameters
    ----------
    variance : float, Tensor
        the variance of the Normal likelihood, this can be made a tf.Variable
        if you want to learn this.

    """

    def __init__(self, variance):
        """Construct an instance of a Normal likelihood."""
        self.variance = variance

    def _loglike(self, y, f):
        """Build the log likelihood.

        Parameters
        ----------
        y : Tensor
            the target variable of shape (N, tasks)
        f : Tensor
            the latent function output from the network of shape (N, tasks)

        """
        ll = -0.5 * (tf.log(2 * self.variance * np.pi) +
                     (y - f)**2 / self.variance)
        return ll


class Bernoulli(Likelihood):
    """Bernoulli log-likelihood."""

    def _loglike(self, y, f):
        """Build the log likelihood.

        Parameters
        ----------
        y : Tensor
            the target variable of shape (N, tasks)
        f : Tensor
            the latent function output from the network of shape (N, tasks)

        """
        ll = y * tf.log(pos(f)) + (1 - y) * tf.log(pos(1 - f))
        return ll


class Categorical(Likelihood):
    """Categorical, or Generalized Bernoulli log-likelihood."""

    def _loglike(self, y, f):
        """Build the log likelihood.

        Parameters
        ----------
        y : Tensor
            the target variable of shape (N, tasks)
        f : Tensor
            the latent function output from the network of shape (N, tasks)

        """
        # sum along last axis, which is assumed to be the `tasks` axis
        ll = tf.reduce_sum(y * tf.log(pos(f)), axis=-1)
        return ll


class Binomial(Likelihood):
    """Binomial log-likelihood.

    Parameters
    ----------
    n : float, ndarray, Tensor
        the number of trials of this binomial distribution

    """

    def __init__(self, n):
        """Construct an instance of a Binomial likelihood."""
        self.n = n

    def _loglike(self, y, f):
        """Build the log likelihood.

        Parameters
        ----------
        y : Tensor
            the target variable of shape (N, tasks)
        f : Tensor
            the latent function output from the network of shape (N, tasks)

        """
        bincoef = tf.lgamma(self.n + 1) - tf.lgamma(y + 1) \
            - tf.lgamma(self.n - y + 1)
        ll = bincoef + y * tf.log(pos(f)) + (self.n - y) * tf.log(pos(1 - f))
        return ll
